// Property automaton -*- c++ -*-

#ifndef PROPERTY_H_
# define PROPERTY_H_
# pragma interface

# include "PropertyState.h"
# include "BitVector.h"

/** @file Property.h
 * Automaton representing the negation of a property being verified
 */

/** Property automaton */
class Property
{
public:
  /** Logical operators */
  enum Op {
    opFalse, opTrue, opNot, opAnd, opOr,
    opFinally, opGlobally, opNext, opUntil, opRelease
  };

  /** Constructor
   * @param printer	Printer object for diagnostics
   */
  explicit Property (const class Printer& printer);
private:
  /** Copy constructor */
  explicit Property (const class Property& old);
  /** Assignment operator */
  class Property operator= (const class Property& old);
public:
  /** Destructor */
  ~Property ();

  /** Create the property automaton from an expression
   * @param translator	command line for invoking an external translator
   * @param expr	expression to be translated to an automaton
   * @return		true if the automaton was successfully generated
   */
  bool create (const char* translator,
	       class Expression& expr);

  /** Append an operator to the formula string
   * @param op		operator to be appended
   */
  void addOp (enum Op op);

  /** Append a proposition to the formula string
   * @param expr	proposition to be appended
   */
  void addExpression (class Expression& expr);

  /** Get the printer object */
  const class Printer& getPrinter () const { return myPrinter; }

  /** Determine the index number of the initial state */
  unsigned getInitialState () const { return myInitialState; }

  /** Determine the number of states in the property automaton */
  unsigned getNumStates () const { return myNumStates; }
  /** Determine the number of acceptance sets in the property automaton */
  unsigned getNumSets () const { return myNumSets; }

  /** Access a state of the automaton
   * @param index	number of the state
   * @return		the state
   */
  const class PropertyState& operator[] (unsigned index) const {
    assert (index < myNumStates);
    return myStates[index];
  }
  /** Determine whether a state belongs to an acceptance set
   * @param index	number of the state
   * @param accept	number of the acceptance set
   */
  bool accepts (unsigned index, unsigned accept) const {
    assert (myStatesAccept && index < myNumStates && accept < myNumSets);
    return (*myStatesAccept)[index * myNumSets + accept];
  }

private:
  /** The Printer object for diagnostic output */
  const class Printer& myPrinter;
  /** File handle for translating the formula */
  int myFD;
  /** Number of atomic propositions */
  unsigned myNumExprs;
  /** The atomic propositions (boolean expressions, the alphabet) */
  class Expression** myExprs;
  /** Number of states */
  unsigned myNumStates;
  /** The states */
  class PropertyState* myStates;
  /** Acceptance sets the states belong to */
  class BitVector* myStatesAccept;
  /** Index number of the initial state */
  unsigned myInitialState;
  /** Number of acceptance sets */
  unsigned myNumSets;
};

#endif // PROPERTY_H_
