// Buffer indexing expression (used only during parsing) -*- c++ -*-

#ifndef BUFFERINDEX_H_
# define BUFFERINDEX_H_
# pragma interface

# include "Expression.h"

/** @file BufferIndex.h
 * Buffer indexing expression (transient storage used during parsing)
 */

/** Buffer indexing expression (used only during parsing) */
class BufferIndex : public Expression
{
public:
  /** Constructor
   * @param buffer	The buffer
   * @param index	Indexing expression
   */
  BufferIndex (class Expression& buffer, class Expression& index);
private:
  /** Copy constructor */
  BufferIndex (const class BufferIndex& old);
  /** Assignment operator */
  class BufferIndex& operator= (const class BufferIndex& old);
protected:
  /** Destructor */
  ~BufferIndex ();
public:

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eBufferIndex; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  /** Get the buffer expression */
  class Expression& getBuffer () { return *myBuffer->copy (); }
  /** Get the indexing expression */
  class Expression& getIndex () { return *myIndex->copy (); }

  /** Equality comparison operator */
  bool operator== (const class BufferIndex& other) const {
    return *myBuffer == *other.myBuffer && *myIndex == *other.myIndex;
  }

  /** Ordering comparison operator */
  bool operator< (const class BufferIndex& other) const {
    if (*myIndex < *other.myIndex) return true;
    if (*other.myIndex < *myIndex) return false;
    return *myBuffer < *other.myBuffer;
  }

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation& valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @param transition	Transition for registering quantified variables
   * @param declare	flag: declare new variables if required
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation,
			    class Transition* transition,
			    bool declare);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

  /** Determine whether the expression depends on a set of variables
   * @param vars	the set of variables
   * @param complement	flag: treat the set as its complement
   */
  bool depends (const class VariableSet& vars,
		bool complement) const;

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @param data	parameters to be passed to the operation
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class Expression&,void*),
		     void* data) const;

# ifdef EXPR_COMPILE
  /** Generate C code for evaluating the expression
   * @param cexpr	the compilation
   * @param indent	indentation level
   * @param lvalue	C expression referring to the lvalue
   * @param vars	the variables that have been assigned a value
   */
  void compile (class CExpression& cexpr,
		unsigned indent,
		const char* lvalue,
		const class VariableSet* vars) const;
# endif // EXPR_COMPILE

  /** Display the expression
   * @param printer	the printer object
   */
  void display (const class Printer& printer) const;

private:
  /** The buffer expression */
  class Expression* myBuffer;
  /** The indexing expression */
  class Expression* myIndex;
};

#endif // BUFFERINDEX_H_
