// Expression list class -*- c++ -*-

#pragma implementation
#include "ExpressionList.h"
#include "Expression.h"
#include <cstring>

/** @file ExpressionList.C
 * List of expressions
 */

ExpressionList::ExpressionList () :
  myReferences (1), mySize (0), myExprs (0)
{
}

ExpressionList::~ExpressionList ()
{
  assert (!myReferences);
  for (card_t i = 0; i < size (); i++)
    myExprs[i]->destroy ();
  delete[] myExprs;
}

void
ExpressionList::append (class Expression& expression)
{
  assert (expression.getKind () == Expression::eUndefined ||
	  expression.getType ());
  if (empty ()) {
    assert (!myExprs);
    myExprs = new class Expression*[1];
  }
  else {
    class Expression** exprs = new class Expression*[mySize + 1];
    memcpy (exprs, myExprs, mySize * sizeof *exprs);
    delete[] myExprs;
    myExprs = exprs;
  }

  myExprs[mySize++] = &expression;
}

void
ExpressionList::prepend (class Expression& expression)
{
  assert (expression.getKind () == Expression::eUndefined ||
	  expression.getType ());
  if (empty ()) {
    assert (!myExprs);
    myExprs = new class Expression*[1];
  }
  else {
    class Expression** exprs = new class Expression*[mySize + 1];
    memcpy (exprs + 1, myExprs, mySize * sizeof *exprs);
    delete[] myExprs;
    myExprs = exprs;
  }

  *myExprs = &expression, mySize++;
}

const class Type*
ExpressionList::getType () const
{
  const class Type* type = NULL;
  for (card_t i = 0; i < mySize; i++)
    if ((type = myExprs[i]->getType ()))
      break;
  return type;
}

bool
ExpressionList::isBasic () const
{
  for (card_t i = 0; i < size (); i++)
    if (!myExprs[i]->isBasic ())
      return false;
  return true;
}

bool
ExpressionList::isTemporal () const
{
  for (card_t i = 0; i < size (); i++)
    if (myExprs[i]->isTemporal ())
      return true;
  return false;
}

bool
ExpressionList::operator== (const class ExpressionList& other) const
{
  if (this == &other)
    return true;
  if (size () != other.size ())
    return false;
  for (card_t i = 0; i < size (); i++)
    if (!(*myExprs[i] == *other.myExprs[i]))
      return false;
  return true;
}

bool
ExpressionList::operator< (const class ExpressionList& other) const
{
  if (this == &other)
    return false;
  if (size () < other.size ())
    return true;
  if (other.size () < size ())
    return false;
  for (card_t i = 0; i < size (); i++) {
    if (*myExprs[i] < *other.myExprs[i])
      return true;
    if (*other.myExprs[i] < *myExprs[i])
      return false;
  }
  return false;
}

class ExpressionList*
ExpressionList::ground (const class Valuation& valuation,
			class Transition* transition,
			bool declare)
{
  bool same = true;
  class ExpressionList* l = new class ExpressionList;

  for (card_t i = 0; i < size (); i++) {
    class Expression* e = myExprs[i]->ground (valuation, transition, declare);
    if (!e) {
      l->destroy ();
      return NULL;
    }

    assert (valuation.isOK ());

    if (e != myExprs[i])
      same = false;
    l->append (*e);
  }

  if (same) {
    l->destroy ();
    return copy ();
  }

  return l;
}

class ExpressionList*
ExpressionList::substitute (class Substitution& substitution)
{
  bool same = true;
  class ExpressionList* l = new class ExpressionList;

  for (card_t i = 0; i < size (); i++) {
    class Expression* e = myExprs[i]->substitute (substitution);
    if (e != myExprs[i])
      same = false;
    l->append (*e);
  }

  if (same) {
    l->destroy ();
    return copy ();
  }

  return l;
}

bool
ExpressionList::depends (const class VariableSet& vars,
			 bool complement) const
{
  for (card_t i = 0; i < size (); i++)
    if (myExprs[i]->depends (vars, complement))
      return true;
  return false;
}

bool
ExpressionList::forVariables (bool (*operation)
			      (const class Expression&,void*),
			      void* data) const
{
  for (card_t i = 0; i < size (); i++)
    if (!myExprs[i]->forVariables (operation, data))
      return false;
  return true;
}
