// Persistent reachability graph storage -*- c++ -*-

#pragma implementation
#include "Graph.h"
#include "LSTS.h"
#include "BTree.h"
#include "Net.h"
#include "Place.h"
#include "Transition.h"
#include "Valuation.h"
#include "VariableDefinition.h"
#include "GlobalMarking.h"
#include "Marking.h"
#include "CardType.h"
#include "LeafValue.h"
#include "BitVector.h"
#include "util.h"

#include <cerrno>
#ifdef SLIST_LOC
# include SLIST_LOC
#else
# include <slist>
#endif

/** @file Graph.C
 * Persistent, lossless reachability graph storage
 */

/** file offset to state directory slot
 * @param s	number of the state
 * @return	file offset of the corresponding directory slot
 */
#define DIRSLOT1(s) (myDirectoryOffset + (s) * (3 * sizeof (fpos_t)))
/** file offset to the second component of a state directory slot
 * @param s	number of the state
 * @return	file offset to the second component of the slot
 */
#define DIRSLOT2(s) (DIRSLOT1 (s) + sizeof (fpos_t))
/** file offset to the third component of a state directory slot
 * @param s	number of the state
 * @return	file offset to the second component of the slot
 */
#define DIRSLOT3(s) (DIRSLOT1 (s) + 2 * sizeof (fpos_t))

/** File offset flag */
#define OFFSET_FLAG LONG_MIN
/** Determine whether an offset has been flagged
 * @param offset	offset to be checked
 * @return		whether the offset has the error bit set
 */
inline static bool
isFlagged (Graph::fpos_t offset)
{
  return offset & OFFSET_FLAG && true;
}

/** Flag an offset
 * @param state		offset to be flagged
 */
inline static void
flag (Graph::fpos_t& offset)
{
  assert (!::isFlagged (offset));
  offset |= OFFSET_FLAG;
}

inline static Graph::fpos_t
getOffset (Graph::fpos_t offset)
{
  return offset & ~OFFSET_FLAG;
}
#undef OFFSET_FLAG

Graph::Graph (const class Net& net) :
  myNet (net), myStateBuf (), myEventBuf (),
  mySucc (new card_t[1]),
  myNumArcs (0), myNumStates (0), myStates (0),
  myDirectory (0), myStateStorage (0), myArcStorage (0), myPredStorage (0),
  myDirectoryOffset (0), myFilename (0), myFilebase (0), myLSTS (0)
{
  *mySucc = 0;
}

Graph::Graph (const class Net& net,
	      const char* filename,
	      const char* filebase) :
  myNet (net), myStateBuf (), myEventBuf (),
  mySucc (new card_t[1]),
  myNumArcs (0), myStates (0),
  myDirectory (0), myStateStorage (0), myArcStorage (0), myPredStorage (0),
  myDirectoryOffset (0), myFilename (0), myFilebase (0), myLSTS (0)
{
  *mySucc = 0;
  unsigned i;
  assert (filename || filebase);
  if (filename)
    myFilename = newString (filename);
  if (filebase)
    myFilebase = newString (filebase);
  else {
    // remove directory component
    for (i = 0; filename[i]; i++);
    while (i > 0 && filename[i - 1] != '/') i--;
    filebase = filename + i;

    // remove the last suffix starting with '.'
    for (i = 0; filebase[i]; i++);
    while (i > 0 && filebase[i] != '.') i--;
    if (!i) // no suffix
      for (i = 0; filebase[i]; i++);
    char* name = new char[i + 1];
    memcpy (name, filebase, i);
    name[i] = 0;
    myFilebase = name;
  }
}

Graph::~Graph ()
{
  if (myDirectory) {
    fclose (myDirectory);
    fclose (myStateStorage);
    fclose (myArcStorage);
    fclose (myPredStorage);
  }

  delete myStates;
  delete[] mySucc;
  delete[] myFilename;
  delete[] myFilebase;
  delete myLSTS;
}

bool
Graph::openFiles (bool regenerate)
{
  assert (!myDirectory && !myStateStorage &&
	  !myArcStorage && !myPredStorage);
  unsigned i;
  for (i = 0; myFilebase[i]; i++);

  // file access modes
  static const char reading[] = "rb+", writing[] = "wb+";
  const char* mode = regenerate ? writing : reading;

  if (myFilename) {
    myDirectory = fopen (myFilename, "rb");
    if (myDirectory)
      fclose (myDirectory);
    else
      return false;
  }

  {
    char* basename = new char[i + 5];
    memcpy (basename, myFilebase, i);

    memcpy (basename + i, ".rgh", 5);
    if (FILE* f = fopen (basename, mode))
      myStates = new class BTree (f);
    else if (errno == ENOENT && myFilename) {
      if (!(f = fopen (basename, mode = writing))) {
	perror (basename);
	delete[] basename;
	return false;
      }
      myStates = new class BTree (f);
    }
    else {
      perror (basename);
      delete[] basename;
      return false;
    }

    memcpy (basename + i, ".rgd", 5);
    if (!(myDirectory = fopen (basename, mode))) {
      perror (basename);
      delete myStates; myStates = 0;
      delete[] basename;
      return false;
    }
    memcpy (basename + i, ".rgs", 5);
    if (!(myStateStorage = fopen (basename, mode))) {
      perror (basename);
      delete myStates; myStates = 0;
      fclose (myDirectory); myDirectory = 0;
      delete[] basename;
      return false;
    }
    memcpy (basename + i, ".rga", 5);
    if (!(myArcStorage = fopen (basename, mode))) {
      perror (basename);
      delete myStates; myStates = 0;
      fclose (myDirectory); myDirectory = 0;
      fclose (myStateStorage); myStateStorage = 0;
      delete[] basename;
      return false;
    }
    memcpy (basename + i, ".rgp", 5);
    if (!(myPredStorage = fopen (basename, mode))) {
      perror (basename);
      delete myStates; myStates = 0;
      fclose (myDirectory); myDirectory = 0;
      fclose (myStateStorage); myStateStorage = 0;
      fclose (myArcStorage); myArcStorage = 0;
      delete[] basename;
      return false;
    }
    delete[] basename;
  }

  static const char magic[] = "Maria Reachability Graph: ";
  const size_t magiclen = (sizeof magic) - 1;

  if (mode == reading) {
    const size_t capacity = 4096;
    char* buf = new char[capacity];
    assert (capacity >= magiclen);

    // check the magic cookie
    if ((fread (buf, 1, magiclen, myDirectory) != magiclen) ||
	memcmp (buf, magic, magiclen)) {
      fprintf (stderr, "%s: graph: wrong magic cookie\n",
	       myFilebase);
    wrongFormat:
      if (feof (myDirectory))
	fprintf (stderr, "%s: premature EOF of graph directory\n",
		 myFilebase);
    wrongFormatNoEOF:
      delete myStates; myStates = 0;
      fclose (myDirectory); myDirectory = 0;
      fclose (myStateStorage); myStateStorage = 0;
      fclose (myArcStorage); myArcStorage = 0;
      fclose (myPredStorage); myPredStorage = 0;
      delete[] buf;
      return false;
    }

    {
      size_t got = fread (buf, 1, capacity, myDirectory);
      size_t len;
      for (len = 0; len < got && buf[len]; len++);
      if (!len) {
	fprintf (stderr, "%s: graph: empty model name\n", myFilebase);
      argError:
	goto wrongFormatNoEOF;
      }
      else if (len == got) {
	fprintf (stderr, "%s: graph: model name is too long\n", myFilebase);
	goto argError;
      }

      if (fseek (myDirectory, fpos_t (len) - got + 1, SEEK_CUR)) {
	fprintf (stderr, "%s: ", myFilebase);
	perror ("fseek");
	goto wrongFormatNoEOF;
      }

      if (myFilename) {
	if (strcmp (buf, myFilename)) {
	  fprintf (stderr, "%s: model names differ: \"%s\" \"%s\"\n",
		   myFilebase, buf, myFilename);
	  goto argError;
	}
      }
      else
	myFilename = newString (buf);
    }

    if (fread (&myNumArcs, sizeof myNumArcs, 1, myDirectory) != 1)
      goto wrongFormat;

    myDirectoryOffset = ftell (myDirectory);
    fseek (myDirectory, 0, SEEK_END);
    fpos_t endpos = ftell (myDirectory);
    if ((endpos - myDirectoryOffset) % (2 * sizeof (fpos_t)))
      goto wrongFormat;
    myNumStates = (endpos - myDirectoryOffset) / (3 * sizeof (fpos_t));
    delete[] buf;
  }
  else { // create the directory afresh
    // magic cookie
    fwrite (magic, magiclen, 1, myDirectory);
    fwrite (myFilename, 1, 1 + strlen (myFilename), myDirectory);
    fwrite (&myNumArcs, sizeof myNumArcs, 1, myDirectory);
    myDirectoryOffset = ftell (myDirectory);
    fflush (myDirectory);
  }

  return true;
}

void
Graph::setLSTS (class LSTS* lsts)
{
  delete myLSTS;
  myLSTS = lsts;
}

bool
Graph::encode (class BitPacker& buf,
	       const class GlobalMarking& m,
	       card_t* errorplace)
{
  // Clear the encoding buffer
  buf.clear ();
  /** Valuation for checking the consistency of implicit places */
  class Valuation valuation;
  valuation.setGlobalMarking (&m);
  // Encode each place marking in the global marking
  for (card_t index = m.getSize (); index--; ) {
    /** Current place marking */
    const class PlaceMarking& place = m[index];
    assert (!!place.getPlace ());
    if (place.getPlace ()->isImplicit ()) {
      const class Marking* m = place.getPlace ()->getInitMarking ();
      assert (!!m);
      class PlaceMarking p;
      p.setPlace (place.getPlace ());
      if (!m->add (p, 1, valuation) || !(p == place)) {
	if (errorplace) *errorplace = index;
	return false;
      }
      continue;
    }

    /** Number of distinct tokens */
    size_t distinct = place.size ();
    /** Array of sorted cardinality,value pairs */
    class PlaceMarking::card_value* histogram =
      distinct ? new class PlaceMarking::card_value[distinct] : 0;
    /** Total cardinality */
    card_t total = distinct ? place.sort (histogram) : 0;

    if (total == CARD_T_MAX) {
    capacity:
      if (errorplace) *errorplace = index;
      delete[] histogram;
      return false;
    }

    // Encode the total number of tokens
    if (const class Constraint* c = place.getPlace ()->getCapacity ()) {
      card_t n;
      if (!CardType::convert (n, total, *c))
	goto capacity;
      if (unsigned cap = place.getPlace ()->getCapacityBits ()) {
	if (cap > 2 && !place.getPlace ()->isNonempty ()) {
	  if (!total) {
	    buf.append (0, 1);
	    delete[] histogram;
	    continue;
	  }
	  buf.append (1, 1);
	}
	buf.append (n, cap);
      }
    }
    else {
      assert (!place.getPlace ()->isNonempty ());
      // variable-length code:
      // 0		0
      // 1..8		10xxx
      // 9..264		110xxxxxxxx
      // 265..65800	1110xxxxxxxxxxxxxxxx
      // 65801..	1111xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
      if (!total)
	buf.append (0, 1);
      else {
	buf.append (1, 1);
	if (total <= 8)
	  buf.append (0, 1), buf.append (total - 1, 3);
	else {
	  buf.append (1, 1);
	  if (total <= 8 + 256)
	    buf.append (0, 1), buf.append (total - (1 + 8), 8);
	  else {
	    buf.append (1, 1);
	    if (total <= 8 + 256 + 65536)
	      buf.append (0, 1), buf.append (total - (1 + 8 + 256), 16);
	    else
	      buf.append (1, 1), buf.append (total, CARD_T_BIT);
	  }
	}
      }
    }

    if (!total) {
      delete[] histogram;
      continue;
    }

    // Encode the number of distinct tokens
    assert (distinct && distinct <= total);
    if (total > 1)
      buf.append (distinct - 1, log2 (total));

    // Encode each distinct token in descending order of multiplicity
    while (distinct--) {
      // Encode the multiplicity (distinct has been decremented by one)
      assert (total >= distinct + 1);
      /** Lower multiplicity bound, inclusive */
      const card_t low = (total + distinct) / (distinct + 1);
      /** Higher multiplicity bound, inclusive */
      const card_t high = total - distinct;
      assert (low <= high);
      // Update the total cardinality of remaining tokens
      assert (total >= histogram[distinct].card);
      total -= histogram[distinct].card;

      assert (histogram[distinct].card >= low &&
	      histogram[distinct].card <= high);
      // Finally encode the multiplicity
      if (high != low)
	buf.append (histogram[distinct].card - low, log2 (high - low + 1));
      // Encode the value
      buf.append (*histogram[distinct].value);
    }

    assert (!total);
    delete[] histogram;
  }

  return true;
}

void
Graph::encode (class BitPacker& buf,
	       const class Net& net,
	       const class Transition& transition,
	       const class Valuation& valuation)
{
  assert (&net.getTransition (transition.getIndex ()) == &transition);
  assert (valuation.isOK ());
  if (net.getNumTransitions () > 1)
    buf.append (transition.getIndex (), log2 (net.getNumTransitions ()));
  else
    assert (net.getNumTransitions () == 1);

  for (Transition::const_iterator i = transition.begin ();
       i != transition.end (); i++) {
    const class Value* value = valuation.getValue (*i->second);
    if (i->second->isUndefined ())
      buf.append (!value, 1);
    else
      assert (!!value);
    if (value)
      buf.append (*value);
  }
}

void
Graph::decode (class GlobalMarking& m,
	       const word_t* data)
{
  card_t index;
  class BitUnpacker buf (data);
  // Decode each place marking in the global marking
  for (index = m.getSize (); index--; ) {
    /** Current place */
    const class Place& place = *m[index].getPlace ();
    if (place.isImplicit ())
      continue;
    /** Total cardinality */
    card_t total;

    // Decode the total number of tokens
    if (const class Constraint* c = place.getCapacity ()) {
      if (unsigned cap = place.getCapacityBits ()) {
	if (cap > 2 && !place.isNonempty () && !buf.extract (1))
	  continue;
	total = CardType::convert (buf.extract (cap), *c);
      }
      else
	total = CardType::convert (0, *c);
    }
    else {
      if (!buf.extract (1))		// 0: 0
	continue;
      else if (!buf.extract (1))	// 10xxx: 1..8
	total = 1 + buf.extract (3);
      else if (!buf.extract (1))	// 110xxxxxxxx: 9..264
	total = (1 + 8) + buf.extract (8);
      else if (!buf.extract (1))	// 1110xxxxxxxxxxxxxxxx: 265..65800
	total = (1 + 8 + 256) + buf.extract (16);
      else {				// 1111...: 65801..
	total = buf.extract (CARD_T_BIT);
	assert (total > 8 + 256 + 65536);
      }
    }

    if (!total)
      continue;

    /** Number of distinct tokens */
    card_t distinct = total > 1 ? buf.extract (log2 (total)) + 1 : 1;

    // Decode each distinct token in descending order of multiplicity
    while (distinct--) {
      // Decode the multiplicity (distinct has been decremented by one)
      assert (total >= distinct + 1);
      /** Lower multiplicity bound, inclusive */
      const card_t low = (total + distinct) / (distinct + 1);
      /** Higher multiplicity bound, inclusive */
      const card_t high = total - distinct;
      assert (low <= high);
      /** Multiplicity */
      const card_t count = low != high
	? buf.extract (log2 (high - low + 1)) + low
	: low;
      // Update the total cardinality of remaining tokens
      assert (total >= count);
      total -= count;

      // Decode the value
      if (!m[index].add (*buf.extract (place.getType ()), count))
	assert (false);
    }

    assert (!total);
  }

  // Compute the markings of implicit places
  class Valuation valuation;
  valuation.setGlobalMarking (&m);
  for (index = m.getSize (); index--; ) {
    const class Place& place = *m[index].getPlace ();
    if (place.isImplicit ()) {
      const class Marking* expr = place.getInitMarking ();
      assert (expr && m[index].empty ());
      if (!expr->add (m[index], 1, valuation))
	assert (false);
    }
  }
}

void
Graph::decode (const class Net& net,
	       class BitUnpacker& buf,
	       const class Transition*& transition,
	       class Valuation& valuation)
{
  transition = &net.getTransition
    (net.getNumTransitions () > 1
     ? buf.extract (log2 (net.getNumTransitions ()))
     : 0);

  for (Transition::const_iterator i = transition->begin ();
       i != transition->end (); i++)
    if (!i->second->isUndefined () || !buf.extract (1))
      valuation.setValue (*i->second, *buf.extract (i->second->getType ()));
}

/** Compute a hash value for a sequence of bytes
 * @param buf		the sequence
 * @param bytes		number of bytes in the sequence
 */
inline static size_t
getHashValue (const unsigned char* buf, size_t bytes)
{
  size_t h = 0;
  while (bytes--)
    h ^= (h << 3 | h >> (CHAR_BIT - 3)) ^ buf[bytes];
  return h;
}

const class Graph::AddStatus
Graph::add (const void* buf,
	    size_t bytes)
{
  size_t h = getHashValue (reinterpret_cast<const unsigned char*>(buf), bytes);
  if (card_t* states = myStates->search (h)) {
    word_t* tmp =
      new word_t[(bytes + (sizeof (word_t) - 1)) / sizeof (word_t)];

    for (card_t s = *states; s; s--) {
      if (update (buf, tmp, bytes, states[s])) {
	s = states[s];
	delete[] tmp;
	delete[] states;
	return AddStatus (false, s);
      }
    }

    delete[] tmp;
    delete[] states;
  }

  // a new state: add it to disk
  {
    fseek (myDirectory, DIRSLOT1 (myNumStates), SEEK_SET);
    if (fseek (myStateStorage, 0, SEEK_END))
      assert (false);
    fpos_t num[3];
    num[0] = ftell (myStateStorage);
    num[1] = num[2] = FPOS_NONE;
    if ((fwrite (&num, sizeof num, 1, myDirectory) != 1) ||
	(fwrite (buf, 1, bytes, myStateStorage) != bytes)) {
      perror ("Graph::add (state): fwrite");
      assert (false);
    }
  }

  myStates->insert (h, myNumStates);
  return AddStatus (true, myNumStates++);
}

const class Graph::AddStatus
Graph::add (const class GlobalMarking& m,
	    card_t* errorplace)
{
  if (encode (myStateBuf, m, errorplace)) {
    myStateBuf.deflate ();
    class AddStatus status =
      add (myStateBuf.getBuf (), myStateBuf.getNumBytes ());
    if (status.isNew && myLSTS) myLSTS->outputState (m, status.number);
    return status;
  }
  else
    return AddStatus (false, CARD_T_MAX);
}

void
Graph::add (card_t source, card_t target)
{
  assert (!!mySucc);
  if (!(*mySucc & (*mySucc + 1))) {
    card_t* succ = new card_t[(*mySucc + 1) << 1];
    memcpy (succ, mySucc, (*mySucc + 1) * sizeof *succ);
    delete[] mySucc;
    mySucc = succ;
  }
  mySucc[++(*mySucc)] = target;
  addReverse (source, target);
}

void
Graph::add (card_t source, card_t target,
	    const class Transition& transition,
	    const class Valuation& valuation)
{
  add (source, target);
  encode (myEventBuf, myNet, transition, valuation);
}

void
Graph::addEvents (card_t source,
		  const void* buf,
		  size_t bytes)
{
  assert (!isProcessed (source) && mySucc);
  /** offset to the arc file */
  fpos_t arcoffset;

  if (*mySucc) {
    if (myLSTS) {
      word_t* data = const_cast<word_t*>(static_cast<const word_t*>(buf));
      BitPacker::inflate (data[(bytes - 1) / sizeof (word_t)],
			  (-bytes) % sizeof (word_t));
      myLSTS->outputArcs (*this, source, mySucc, data);
      BitPacker::deflate (data[(bytes - 1) / sizeof (word_t)],
			  (-bytes) % sizeof (word_t));
    }
    fseek (myArcStorage, 0, SEEK_END);
    arcoffset = ftell (myArcStorage);
    /** number of successor states and length of the encoded data */
    card_t bytecode = bytes << 16 | (*mySucc &= 0xffff);

    if ((1 != fwrite (&bytecode, sizeof bytecode, 1, myArcStorage)) ||
	(*mySucc != fwrite (mySucc + 1, sizeof *mySucc,
			    *mySucc, myArcStorage)) ||
	(bytes != fwrite (buf, 1, bytes, myArcStorage))) {
      perror ("Graph::add (event): fwrite");
      assert (false);
    }
    *mySucc = 0;
  }
  else
    arcoffset = FPOS_DEAD;

  // update the directory file

  fseek (myDirectory, DIRSLOT2 (source), SEEK_SET);
#ifndef NDEBUG
  {
    fpos_t arcpos;
    assert (1 == fread (&arcpos, sizeof arcpos, 1, myDirectory) &&
	    arcpos == FPOS_NONE);
    fseek (myDirectory, DIRSLOT2 (source), SEEK_SET);
  }
#endif // NDEBUG
  fwrite (&arcoffset, sizeof arcoffset, 1, myDirectory);
  fseek (myDirectory, myDirectoryOffset - sizeof myNumArcs, SEEK_SET);
  fwrite (&myNumArcs, sizeof myNumArcs, 1, myDirectory);
}

void
Graph::addEvents (card_t source)
{
  myEventBuf.deflate ();
  addEvents (source, myEventBuf.getBuf (), myEventBuf.getNumBytes ());
  myEventBuf.clear ();
}

word_t*
Graph::fetchState (card_t number, unsigned& length, bool* erroneous) const
{
  assert (number < myNumStates);
  fseek (myDirectory, DIRSLOT1 (number), SEEK_SET);
  fpos_t num;
  if (fread (&num, sizeof num, 1, myDirectory) != 1) {
  readError:
    perror ("Graph::fetchState (): fread");
    return 0;
  }
  if (erroneous)
    *erroneous = ::isFlagged (num);
  num = ::getOffset (num);
  fpos_t pos;
  if (number == myNumStates - 1) {
    fseek (myStateStorage, 0, SEEK_END);
    length = ftell (myStateStorage) - num;
  }
  else {
    fseek (myDirectory, DIRSLOT1 (number + 1), SEEK_SET);
    if (fread (&pos, sizeof pos, 1, myDirectory) != 1)
      goto readError;
    length = ::getOffset (pos) - num;
  }

  if (fseek (myStateStorage, num, SEEK_SET)) {
    perror ("Graph::fetchState (): fseek");
    return 0;
  }
  unsigned numWords = (length + sizeof (word_t) - 1) / sizeof (word_t);
  word_t* buf = new word_t[numWords];
  if (fread (buf, 1, length, myStateStorage) != length) {
    delete[] buf;
    goto readError;
  }
  return buf;
}

class GlobalMarking*
Graph::fetchState (card_t number) const
{
  unsigned length;
  bool erroneous;
  word_t* buf = fetchState (number, length, &erroneous);
  if (!buf)
    return 0;

  BitPacker::inflate (buf[(length - 1) / sizeof (word_t)],
		      (-length) % sizeof (word_t));
  // Decode the marking
  class GlobalMarking* marking = new class GlobalMarking (myNet);
  decode (*marking, buf);
  // Set the error bit
  if (erroneous)
    marking->flagErroneous ();
  delete[] buf;
  return marking;
}

card_t
Graph::getNumPredecessors (card_t state) const
{
  card_t num = 0;
  for (fpos_t pos = getPredecessors (state); pos != FPOS_NONE;
       getPredecessor (pos))
    num++;
  return num;
}

Graph::fpos_t
Graph::getPredecessors (card_t state) const
{
  assert (state < myNumStates);
  fseek (myDirectory, DIRSLOT3 (state), SEEK_SET);
  fpos_t pos;
  if (fread (&pos, sizeof pos, 1, myDirectory) != 1) {
    assert (false);
    return FPOS_NONE;
  }
  return pos;
}

card_t
Graph::getPredecessor (Graph::fpos_t& pos) const
{
  assert (pos != FPOS_NONE);
  fseek (myPredStorage, pos, SEEK_SET);
  card_t state;
  if (fread (&state, sizeof state, 1, myPredStorage) != 1 ||
      fread (&pos, sizeof pos, 1, myPredStorage) != 1) {
    assert (false);
    pos = FPOS_NONE;
    return CARD_T_MAX;
  }
  assert (state < myNumStates);
  return state;
}

card_t*
Graph::getSuccessors (card_t state, word_t** data) const
{
  assert (state < myNumStates);
  fseek (myDirectory, DIRSLOT2 (state), SEEK_SET);
  fpos_t pos;
  if (fread (&pos, sizeof pos, 1, myDirectory) != 1) {
    assert (false);
    return 0;
  }
  if (pos == FPOS_DEAD || pos == FPOS_NONE)
    return 0;
  fseek (myArcStorage, pos, SEEK_SET);
  /** number of successor states and length of the encoded data */
  card_t bytecode;
  if (fread (&bytecode, sizeof bytecode, 1, myArcStorage) != 1) {
    assert (false);
    return 0;
  }
  assert ((bytecode & 0xffff) > 0);
  card_t* succ = new card_t[(bytecode & 0xffff) + 1];
  *succ = bytecode & 0xffff;
  if (fread (succ + 1, sizeof *succ, *succ, myArcStorage) != *succ) {
    assert (false);
    delete[] succ;
    return 0;
  }
  if (data) {
    const size_t length = bytecode >> 16;
    if (!length)
      *data = 0;
    else {
      const unsigned numWords =
	(length + sizeof (word_t) - 1) / sizeof (word_t);
      word_t* buf = new word_t[numWords];
      if (fread (buf, 1, length, myArcStorage) != length) {
	assert (false);
	delete[] succ;
	delete[] buf;
	return 0;
      }
      BitPacker::inflate (buf[numWords - 1], (-length) % sizeof (word_t));
      *data = buf;
    }
  }
  return succ;
}

void
Graph::getSuccessor (class BitUnpacker& buf,
		     const class Transition*& transition,
		     class Valuation& valuation) const
{
  decode (myNet, buf, transition, valuation);
}

#ifdef EXPR_COMPILE
# include "Compilation.h"

void
Graph::fetchState (const class Compilation& compilation,
		   card_t number) const
{
  unsigned length;
  word_t* buf = fetchState (number, length, 0);
  assert (!!buf);
  compilation.stateDecode (buf, length);
  delete[] buf;
}

#endif // EXPR_COMPILE

bool
Graph::eval (card_t state,
	     const class Expression* cond) const
{
  if (!cond)
    return true;
  class GlobalMarking* m = fetchState (state);
  class Valuation valuation;
  bool accept = false;
  valuation.setGlobalMarking (m);
  if (class Value* value = cond->eval (valuation)) {
    assert (value->getKind () == Value::vLeaf);
    assert (valuation.isOK ());
    accept = bool (static_cast<class LeafValue&>(*value));
    delete value;
  }
  delete m;
  return accept;
}

card_t*
Graph::toPath (const PathMap& tree,
	       card_t dest)
{
  card_t* path = new card_t[1];
  *path = 0;
  for (;;) {
    card_t* npath = new card_t[++(*path) + 1];
    memcpy (npath, path, *path * sizeof *path);
    delete[] path; path = npath;
    path[*path] = dest;
    PathMap::const_iterator i = tree.find (dest);
    if (i == tree.end ()) break;
    dest = i->second;
  }
  return path;
}

card_t*
Graph::path (card_t state, card_t target,
	     const class Expression* pathc) const
{
  if (!eval (state, pathc) || !eval (target, pathc))
    return 0;
  /** Visited states */
  class BitVector visited (myNumStates);
  visited.assign (state, true);
  /** Search queue */
  std::slist<card_t> sq;
  sq.push_front (CARD_T_MAX); /* mark for the next level */
  /** Tail of the search queue */
  std::slist<card_t>::iterator sq_tail = sq.begin ();
  sq.insert_after (sq_tail, state), sq_tail++;
  /** Arcs in the explored graph (target, source) */
  PathMap path;

  for (;;) {
    if ((state = *sq.begin ()) == CARD_T_MAX) {
      sq.pop_front ();
      if (sq.empty ())
	return 0;
      sq.insert_after (sq_tail, CARD_T_MAX), sq_tail++;
      state = *sq.begin ();
    }
    sq.pop_front ();
    assert (!sq.empty ());
    assert (visited[state]);
    if (state == target)
      return toPath (path, target);
    if (card_t* succ = getSuccessors (state)) {
      for (assert (*succ > 0); *succ; (*succ)--) {
	card_t s = succ[*succ];
	if (!visited.tset (s) && eval (s, pathc)) {
	  sq.insert_after (sq_tail, s), sq_tail++;
	  path.insert (PathMap::value_type (s, state));
	}
      }
      delete[] succ;
    }
  }
}

card_t*
Graph::path (card_t state, const class Expression& cond,
	     const class Expression* pathc) const
{
  if (!eval (state, pathc))
    return 0;
  /** Visited states */
  class BitVector visited (myNumStates);
  visited.assign (state, true);
  /** Search queue */
  std::slist<card_t> sq;
  sq.push_front (CARD_T_MAX); /* mark for the next level */
  /** Tail of the search queue */
  std::slist<card_t>::iterator sq_tail = sq.begin ();
  sq.insert_after (sq_tail, state), sq_tail++;
  /** Arcs in the explored graph (target, source) */
  PathMap path;

  for (;;) {
    if ((state = *sq.begin ()) == CARD_T_MAX) {
      sq.pop_front ();
      if (sq.empty ())
	return 0;
      sq.insert_after (sq_tail, CARD_T_MAX), sq_tail++;
      state = *sq.begin ();
    }
    sq.pop_front ();
    assert (!sq.empty ());
    assert (visited[state]);
    if (eval (state, &cond))
      return toPath (path, state);
    if (card_t* succ = getSuccessors (state)) {
      for (assert (*succ > 0); *succ; (*succ)--) {
	card_t s = succ[*succ];
	if (!visited.tset (s) && eval (s, pathc)) {
	  sq.insert_after (sq_tail, s), sq_tail++;
	  path.insert (PathMap::value_type (s, state));
	}
      }
      delete[] succ;
    }
  }
}

card_t*
Graph::rpath (card_t state, const class Expression& cond,
	      const class Expression* pathc) const
{
  if (!eval (state, pathc))
    return 0;
  /** Visited states */
  class BitVector visited (myNumStates);
  visited.assign (state, true);
  /** Search queue */
  std::slist<card_t> sq;
  sq.push_front (CARD_T_MAX); /* mark for the next level */
  /** Tail of the search queue */
  std::slist<card_t>::iterator sq_tail = sq.begin ();
  sq.insert_after (sq_tail, state), sq_tail++;
  /** Arcs in the explored graph (target, source) */
  PathMap path;

  for (;;) {
    if ((state = *sq.begin ()) == CARD_T_MAX) {
      sq.pop_front ();
      if (sq.empty ())
	return 0;
      sq.insert_after (sq_tail, CARD_T_MAX), sq_tail++;
      state = *sq.begin ();
    }
    sq.pop_front ();
    assert (!sq.empty ());
    assert (visited[state]);
    if (eval (state, &cond))
      return toPath (path, state);
    for (fpos_t pos = getPredecessors (state); pos != FPOS_NONE;) {
      card_t s = getPredecessor (pos);
      if (!visited.tset (s) && eval (s, pathc)) {
	sq.insert_after (sq_tail, s), sq_tail++;
	path.insert (PathMap::value_type (s, state));
      }
    }
  }
}

card_t*
Graph::path (card_t state,
	     const card_t* loop,
	     const class Expression* pathc) const
{
  if (!eval (state, pathc))
    return 0;
  /** Visited states */
  class BitVector visited (myNumStates);
  visited.assign (state, true);
  /** Search queue */
  std::slist<card_t> sq;
  sq.push_front (CARD_T_MAX); /* mark for the next level */
  /** Tail of the search queue */
  std::slist<card_t>::iterator sq_tail = sq.begin ();
  sq.insert_after (sq_tail, state), sq_tail++;
  /** Arcs in the explored graph (target, source) */
  PathMap path;

  for (;;) {
    if ((state = *sq.begin ()) == CARD_T_MAX) {
      sq.pop_front ();
      if (sq.empty ())
	return 0;
      sq.insert_after (sq_tail, CARD_T_MAX), sq_tail++;
      state = *sq.begin ();
    }
    sq.pop_front ();
    assert (!sq.empty ());
    assert (visited[state]);
    for (const card_t* l = loop + *loop; l > loop; l--) {
      if (state == *l) {
	card_t* p = toPath (path, state);
	// append the loop to the path
	card_t* q = new card_t[*p + *loop];
	*q = *p + *loop - 1;
	memcpy (q + *loop, p + 1, *p * sizeof *p);
	delete[] p;
	p = q + *loop;
	const card_t* k = l;
	for (; l++ < loop + *loop; *--p = *l);
	for (l = loop + 1; l++ < k; *--p = *l);
	assert (p == q + 1);
	return q;
      }
    }
    if (card_t* succ = getSuccessors (state)) {
      for (assert (*succ > 0); *succ; (*succ)--) {
	card_t s = succ[*succ];
	if (!visited.tset (s) && eval (s, pathc)) {
	  sq.insert_after (sq_tail, s), sq_tail++;
	  path.insert (PathMap::value_type (s, state));
	}
      }
      delete[] succ;
    }
  }
}

bool
Graph::flagErroneous (card_t number)
{
  assert (number < myNumStates);
  fseek (myDirectory, DIRSLOT1 (number), SEEK_SET);
  fpos_t num;
  if (fread (&num, sizeof num, 1, myDirectory) != 1)
    assert (false);
  if (::isFlagged (num))
    return true;
  fseek (myDirectory, -sizeof num, SEEK_CUR);
  ::flag (num);
  if (fwrite (&num, sizeof num, 1, myDirectory) != 1)
    assert (false);
  return false;
}

bool
Graph::isErroneous (card_t number) const
{
  assert (number < myNumStates);
  fseek (myDirectory, DIRSLOT1 (number), SEEK_SET);
  fpos_t num;
  if (fread (&num, sizeof num, 1, myDirectory) != 1)
    assert (false);
  return ::isFlagged (num);
}

bool
Graph::isProcessed (card_t number) const
{
  assert (number < myNumStates);
  fseek (myDirectory, DIRSLOT2 (number), SEEK_SET);
  fpos_t num;
  if (fread (&num, sizeof num, 1, myDirectory) != 1)
    assert (false);
  return num != FPOS_NONE;
}

bool
Graph::update (const void* buf, void* tmp, size_t bytes, card_t number)
{
  fseek (myDirectory, DIRSLOT1 (number), SEEK_SET);
  fpos_t num;
  if (fread (&num, sizeof num, 1, myDirectory) != 1)
    assert (false);

  if (fseek (myStateStorage, ::getOffset (num), SEEK_SET)) {
    perror ("Graph::update (): fseek");
    assert (false);
  }

  return
    fread (tmp, 1, bytes, myStateStorage) == bytes &&
    !memcmp (tmp, buf, bytes);
}

void
Graph::addReverse (card_t source, card_t target)
{
  myNumArcs++;
  fpos_t arcoffset;
  fseek (myDirectory, DIRSLOT3 (target), SEEK_SET);
  if (1 != fread (&arcoffset, sizeof arcoffset, 1, myDirectory)) {
    perror ("Graph::addReverse (): fread");
    assert (false);
    return;
  }
  fseek (myDirectory, -sizeof arcoffset, SEEK_CUR);
  fseek (myPredStorage, 0, SEEK_END);
  fpos_t arcend = ftell (myPredStorage);
  fwrite (&arcend, sizeof arcend, 1, myDirectory);
  fwrite (&source, sizeof source, 1, myPredStorage);
  fwrite (&arcoffset, sizeof arcoffset, 1, myPredStorage);
}
