// Persistent reachability graph storage -*- c++ -*-

#ifndef GRAPH_H_
# define GRAPH_H_
# pragma interface

# include "BitBuffer.h"

# include <map>
# include <cassert>
# include <cstdio>

/** @file Graph.h
 * Persistent, lossless reachability graph storage
 */

/** Persistent reachability graph storage */
class Graph
{
public:
  /** File offset */
  typedef long fpos_t;
  /** nonexistent file offset */
#define FPOS_NONE (-1L)
  /** special offset indicating that a deadlock state */
#define FPOS_DEAD (-2L)

  /** Addition status */
  class AddStatus
  {
  public:
    /** Constructor
     * @param isNew_	flag: was the state just added to the graph?
     * @param number_	the state number
     */
    AddStatus (bool isNew_, card_t number_) :
      isNew (isNew_), number (number_) {
    }
    /** Copy constructor */
    AddStatus (const class AddStatus& old) :
      isNew (old.isNew), number (old.number) {
    }
    /** Assignment operator */
    class AddStatus& operator= (const class AddStatus& old) {
      isNew = old.isNew, number = old.number;
      return *this;
    }
    /** Destructor */
    ~AddStatus () {}
    /** flag: was the state just added to the graph? */
    bool isNew;
    /** the state number */
    card_t number;
  };

  /** Constructor
   * @param net		the net whose reachability graph this is
   */
  explicit Graph (const class Net& net);
  /** Constructor
   * @param net		the net whose reachability graph this is
   * @param filename	name of the Petri Net file (NULL=read from graph file)
   * @param filebase	base name for the graph files (generated if NULL)
   */
  explicit Graph (const class Net& net,
		  const char* filename, const char* filebase);
private:
  /** Copy constructor */
  explicit Graph (const class Graph& old);
  /** Assignment operator */
  class Graph& operator= (const class Graph& old);
public:
  /** Destructor */
  ~Graph ();

  /** Determine the net whose reachability graph this is */
  const class Net& getNet () const { return myNet; }
  /** Determine the number of arcs generated */
  unsigned getNumArcs () const { return myNumArcs; }

  /** Determine the name of the Petri Net */
  const char* getFilename () const { return myFilename; }
  /** Determine the base name for the graph files */
  const char* getFilebase () const { return myFilebase; }

  /** Open the reachability graph files
   * @param regenerate	flag: true=ignore existing graph files
   * @return		true if the files could be opened; false on error
   */
  bool openFiles (bool regenerate);

  /** Set the LSTS output handler
   * @param lsts	the output handler (NULL=none)
   */
  void setLSTS (class LSTS* lsts);
  /** Get the LSTS output handler */
  class LSTS* getLSTS () const { return myLSTS; }

  /** Add the currently encoded state to the graph
   * @param buf		the encoded state
   * @param bytes	size of the encoded state in bytes
   * @return		{false,number} if the state already exists
   *			{true,number} if the state was added
   */
  const class AddStatus add (const void* buf,
			     size_t bytes);

  /** Add a state to the graph
   * @param m		state (marking) to be added
   * @param errorplace	placeholder for the identity of an erroneous place
   * @return		{false,number} if the state already exists
   *			{false,CARD_T_MAX} in case of integrity violation
   *			{true,number} if the state was added
   */
  const class AddStatus add (const class GlobalMarking& m,
			     card_t* errorplace);

  /** Encode a marking
   * @param buf		buffer to receive the marking
   * @param m		marking to be encoded to the buffer
   * @param errorplace	placeholder for the identity of an erroneous place
   * @return		false in case of an integrity violation
   */
  static bool encode (class BitPacker& buf,
		      const class GlobalMarking& m,
		      card_t* errorplace);

  /** Encode a transition instance
   * @param buf		buffer for the encoded data
   * @param net		the high-level Petri Net
   * @param transition	the high-level transition
   * @param valuation	variable definitions
   */
  static void encode (class BitPacker& buf,
		      const class Net& net,
		      const class Transition& transition,
		      const class Valuation& valuation);

  /** Decode a marking
   * @param m		the marking to receive the tokens
   * @param data	the encoded representation
   * @return		the decoded marking
   */
  static void decode (class GlobalMarking& m,
		      const word_t* data);


  /** Decode a transition instance
   * @param net		the high-level Petri Net
   * @param buf		the encoded data buffer
   * @param transition	(output parameter) the transition
   * @param valuation	(output paremeter) the variable bindings
   */
  static void decode (const class Net& net,
		      class BitUnpacker& buf,
		      const class Transition*& transition,
		      class Valuation& valuation);

  /** Add an event to the graph
   * @param source	the source state number
   * @param target	the target state number
   */
  void add (card_t source, card_t target);
  /** Add an event to the graph
   * @param source	the source state number
   * @param target	the target state number
   * @param transition	the transition
   * @param valuation	the binding of the transition
   */
  void add (card_t source, card_t target,
	    const class Transition& transition,
	    const class Valuation& valuation);

  /** Update added events to the arc file and flag the source state processed
   * @param source	the source state number
   * @param buf		the encoded events
   * @param bytes	size of the encoded events in bytes
   */
  void addEvents (card_t source,
		  const void* buf,
		  size_t bytes);
  /** Update added events to the arc file and flag the source state processed
   * @param source	the source state number
   */
  void addEvents (card_t source);

  /** Get the number of states in the reachability graph */
  size_t getNumStates () const { return myNumStates; }

private:
  /** Fetch an encoded state from the reachability graph
   * @param number	number of the state
   * @param length	(output) length of the encoded state in bytes
   * @param erroneous	(output) flag: erroneous state
   * @return		the encoded state, or NULL
   */
  word_t* fetchState (card_t number, unsigned& length, bool* erroneous) const;
public:

  /** Get a state by number
   * @param number	number of the state
   * @return		the decoded state
   */
  class GlobalMarking* fetchState (card_t number) const;

  /** Get the number of predecessor states of a state
   * @param state	number of the state
   * @return		number of directly preceding states
   */
  card_t getNumPredecessors (card_t state) const;

  /** Fetch an offset to the predecessor records for a state
   * @param state	number of the state 
   * @return		offset to the event storage (FPOS_NONE if none)
   */
  fpos_t getPredecessors (card_t state) const;

  /** Fetch predecessor state numbers for a state
   * @param pos		(in/out) offset to the event storage
   * @return		number of the source state (pos==FPOS_NONE at end)
   */
  card_t getPredecessor (fpos_t& pos) const;

  /** Fetch the successor arcs for a state
   * @param state	number of the state
   * @param data	(output) the encoded successors
   * @return		amount and numbers of successor states
   */
  card_t* getSuccessors (card_t state, word_t** data = 0) const;

  /** Decode a successor arc
   * @param buf		the encoded successors
   * @param transition	(output parameter) the transition
   * @param valuation	(output parameter) the variable bindings
   */
  void getSuccessor (class BitUnpacker& buf,
		     const class Transition*& transition,
		     class Valuation& valuation) const;

# ifdef EXPR_COMPILE
  /** Get a state by number
   * @param compilation	the compilation
   * @param number	number of the state
   * @return		the decoded state
   */
  void fetchState (const class Compilation& compilation,
		   card_t number) const;
# endif // EXPR_COMPILE

  /** Evaluate a condition in a state
   * @param state	the state
   * @param cond	the condition (NULL=true)
   * @return		whether the condition holds
   */
  bool eval (card_t state,
	     const class Expression* cond) const;

  /** Tree of paths */
  typedef std::map<card_t,card_t> PathMap;

  /** Extract a path from a leaf to the root of a tree
   * @param tree	the tree
   * @param leaf	the leaf of the tree
   * @return		the path
   */
  static card_t* toPath (const PathMap& tree,
			 card_t dest);

  /** Determine the shortest path from a state to a state
   * @param state	the source state
   * @param target	the target state
   * @param pathc	condition that must hold in every state on the path
   * @return		the amount and numbers of the states on the path
   */
  card_t* path (card_t state, card_t target,
		const class Expression* pathc) const;

  /** Determine the shortest path to a state fulfilling a condition
   * @param state	the source state
   * @param cond	the condition
   * @param pathc	condition that must hold in every state on the path
   * @return		the amount and numbers of the states on the path
   */
  card_t* path (card_t state, const class Expression& cond,
		const class Expression* pathc) const;

  /** Determine the shortest path from a state fulfilling a condition
   * @param state	the target state
   * @param cond	the condition
   * @param pathc	condition that must hold in every state on the path
   * @return		the amount and numbers of the states on the path
   */
  card_t* rpath (card_t state, const class Expression& cond,
		 const class Expression* pathc) const;

  /** Determine the shortest path to a loop state
   * @param state	the source state
   * @param loop	the amount and numbers of the states forming the loop
   * @param cond	the condition
   * @param pathc	condition that must hold in every state on the path
   * @return		the amount and numbers of the states on the path
   */
  card_t* path (card_t state,
		const card_t* loop,
		const class Expression* pathc) const;

  /** Flag a state erroneous
   * @param number	the state number
   * @return		whether the state already was flagged erroneous
   */
  bool flagErroneous (card_t number);

  /** Determine whether a state is erroneous
   * @param number	the state number
   */
  bool isErroneous (card_t number) const;

  /** Determine whether a state has been processed
   * @param number	the state number
   */
  bool isProcessed (card_t number) const;

private:
  /** Find and update an encoded state in the graph
   * @param buf		the encoded state to find or update
   * @param tmp		work storage for encoded states
   * @param bytes	length of the encoded state in bytes
   * @param number	number of the state to be searched
   * @return		whether the state was found
   */
  bool update (const void* buf, void* tmp, size_t bytes, card_t number);
  /** Add a reverse arc to the graph
   * @param source	source state number
   * @param target	target state number
   */
  void addReverse (card_t source, card_t target);

private:
  /** The net whose reachability graph this is */
  const class Net& myNet;
  /** Buffer for encoding states */
  class BitPacker myStateBuf;
  /** Buffer for encoding events */
  class BitPacker myEventBuf;
  /** Buffer for successor state numbers of events */
  card_t* mySucc;
  /** Number of arcs */
  unsigned myNumArcs;
  /** Number of states */
  unsigned myNumStates;
  /** The state hash */
  class BTree* myStates;
  /** The graph directory file */
  FILE* myDirectory;
  /** The state storage file */
  FILE* myStateStorage;
  /** The arc storage file */
  FILE* myArcStorage;
  /** The predecessor state file */
  FILE* myPredStorage;
  /** Offset of the state directory in myDirectory */
  fpos_t myDirectoryOffset;
  /** Name of the Petri Net */
  const char* myFilename;
  /** Base name for the graph files */
  const char* myFilebase;
  /** Labelled state transition system output for newly generated states */
  class LSTS* myLSTS;
};

#endif // GRAPH_H_
