// Place class -*- c++ -*-

#ifndef PLACE_H_
# define PLACE_H_
# pragma interface

# include <cassert>
# include "typedefs.h"

/** @file Place.h
 * Place in an algebraic system net
 */

/** Place class */
class Place
{
  /** The constructor can only be called by Net */
  friend class Net;
protected:
  /** Constructor
   * @param index	Index number of the place (ask Net)
   * @param name	Name of the place
   * @param capacity	Capacity constraint
   * @param type	Type of the place
   */
  Place (unsigned index, char* name,
	 class Constraint* capacity,
	 const class Type& type);
private:
  /** Copy constructor */
  explicit Place (const class Place& old);
  /** Assignment operator */
  class Place& operator= (const class Place& old);
public:
  /** Destructor */
  ~Place ();

  /** Get the index number of the place */
  unsigned getIndex () const { return myIndex; }
  /** Get the name of the place */
  const char* getName () const { return myName; }
  /** Get the type of the place */
  const class Type& getType () const { return *myType; }

  /** Determine the capacity constraint */
  const class Constraint* getCapacity () const { return myCapacity; }
  /** Determine the number of bits needed for storing the capacity */
  unsigned getCapacityBits () const { return myCapacityBits; }

  /** Get the initial marking expression */
  const class Marking* getInitMarking () const { return myInitMarking; }
  /** Set the initial marking expression */
  void setInitMarking (class Marking& initMarking);
  /** Determine whether this is an implicit place */
  bool isImplicit () const { return myIsImplicit; }
  /** Flag the place visible */
  void flagVisible () { myIsVisible = true; }
  /** Determine whether this place is used in deadlock or reject formulae */
  bool isVisible () const { return myIsVisible; }
  /** Determine whether the place is always nonempty */
  bool isNonempty () const { return myIsNonempty; }
  /** Determine whether the place is suppressed when displaying markings */
  bool isSuppressed () const { return myIsSuppressed; }
  /** Specify whether the place is suppressed when displaying markings */
  void setSuppressed (bool suppressed) const { myIsSuppressed = suppressed; }
  /** Determine the maximum number of tokens the place can contain */
  card_t getMaxNumTokens () const { return myMaxNumTokens; }

  /** Display the place definition
   * @param printer	the printer object
   */
  void display (const class Printer& printer) const;

private:
  /** Index number of the place */
  unsigned myIndex;
  /** Name of the place */
  char* myName;
  /** Capacity constraint */
  class Constraint* myCapacity;
  /** Number of bits required to store the capacity */
  unsigned myCapacityBits;
  /** Type of the place */
  const class Type* myType;
  /** Initial marking of the place */
  class Marking* myInitMarking;
  /** Flag: is this an implicit place */
  bool myIsImplicit;
  /** Flag: is this place used in deadlock or reject formulae */
  bool myIsVisible;
  /** Flag: is this place always nonempty */
  bool myIsNonempty;
  /** Flag: is this place suppressed when displaying markings */
  mutable bool myIsSuppressed;
  /** Number of tokens always contained in this place (0 if variable) */
  card_t myMaxNumTokens;
};

#endif // PLACE_H_
